"""
Control and add subsystems to the running daemon hub
"""
import os
from typing import Any
from typing import Callable
from typing import Generator
from typing import List
from typing import Tuple

import pop.hub


def add(
    hub: pop.hub.Hub,
    pypath: List[str] or str = None,
    subname: str = None,
    sub: pop.hub.Sub = None,
    static: List[str] or str = None,
    contracts_pypath: List[str] or str = None,
    contracts_static: List[str] or str = None,
    default_contracts: List[str] or str = None,
    virtual: bool = True,
    dyne_name: str = None,
    omit_start: Tuple[str] = ("_",),
    omit_end: Tuple[str] = (),
    omit_func: bool = False,
    omit_class: bool = True,
    omit_vars: bool = False,
    mod_basename: str = "pop.sub",
    stop_on_failures: bool = False,
    init: bool = True,
    load_all: bool = True,
    recursive_contracts_static: List[str] or str = None,
    default_recursive_contracts: List[str]
    or str = None,  # TODO: Not str, pretty sure -W. Werner, 2020-10-20
):
    """
    Add a new subsystem to the hub
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    :param sub: The sub to use as the root to add to
    :param pypath: One or many python paths which will be imported
    :param static: Directories that can be explicitly passed
    :param contracts_pypath: Load additional contract paths
    :param contracts_static: Load additional contract paths from a specific directory
    :param default_contracts: Specifies that a specific contract plugin will be applied as a default to all plugins
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param dyne_name: The dynamic name to use to look up paths to find plugins -- linked to conf.py
    :param omit_start: Allows you to pass in a tuple of characters that would omit the loading of any object
        I.E. Any function starting with an underscore will not be loaded onto a plugin
        (You should probably never change this)
    :param omit_end:Allows you to pass in a tuple of characters that would omit the loading of an object
        (You should probably never change this)
    :param omit_func: bool: Don't load any functions
    :param omit_class: bool: Don't load any classes
    :param omit_vars: bool: Don't load any vars
    :param mod_basename: str: Manipulate the location in sys.modules that the plugin will be loaded to.
        Allow plugins to be loaded into a separate namespace.
    :param stop_on_failures: If any module fails to load for any reason, stacktrace and do not continue loading this sub
    :param init: bool: determine whether or not we process __init__ functions
    :param load_all: Load all the plugins on the sub
    """
    if pypath:
        pypath = pop.hub.ex_path(pypath)
        subname = subname if subname else pypath[0].split(".")[-1]
    elif static:
        subname = subname if subname else os.path.basename(static)
    if dyne_name:
        subname = subname if subname else dyne_name
    root = sub or hub
    # The dynamic namespace is already on the hub
    if dyne_name in root._subs:
        return
    root._subs[subname] = pop.hub.Sub(
        hub,
        subname,
        root,
        pypath,
        static,
        contracts_pypath,
        contracts_static,
        default_contracts,
        virtual,
        dyne_name,
        omit_start,
        omit_end,
        omit_func,
        omit_class,
        omit_vars,
        mod_basename,
        stop_on_failures,
        init=init,
        sub_virtual=getattr(root, "_subvirt", True),
        recursive_contracts_static=recursive_contracts_static,
        default_recursive_contracts=default_recursive_contracts,
    )
    # init the sub (init.py:__init__) after it can be referenced on the hub!
    root._subs[subname]._sub_init()
    if load_all:
        root._subs[subname]._load_all()
    for alias in root._subs[subname]._alias:
        root._sub_alias[alias] = subname


def remove(hub: pop.hub.Hub, subname: str):
    """
    Remove a pop from the hub, run the shutdown if needed
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    """
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        if hasattr(sub, "init"):
            mod = getattr(sub, "init")
            if hasattr(mod, "shutdown"):
                mod.shutdown()
        hub._remove_subsystem(subname)


def load_all(hub: pop.hub.Hub, subname: str) -> bool:
    """
    Load all modules under a given pop
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    """
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        sub._load_all()
        return True
    else:
        return False


def get_dirs(hub: pop.hub.Hub, sub: pop.hub.Sub) -> List[str]:
    """
    Return a list of directories that contain the modules for this subname
    :param hub: The redistributed pop central hub
    :param sub: The pop object that contains the loaded module data
    """
    return sub._dirs


def iter_subs(
    hub: pop.hub.Hub, sub: pop.hub.Sub, recurse: bool = False
) -> Generator[pop.hub.Sub, None, None]:
    """
    Return an iterator that will traverse just the subs. This is useful for
    nested subs
    :param hub: The redistributed pop central hub
    :param recurse: Recursively iterate over nested subs
    """
    for name in sorted(sub._subs):
        ret = sub._subs[name]
        if ret._sub_virtual:
            yield ret
            if recurse:
                if hasattr(ret, "_subs"):
                    yield from hub.pop.sub.iter_subs(ret, recurse)


def load_subdirs(hub: pop.hub.Hub, sub: pop.hub.Sub, recurse: bool = False):
    """
    Given a sub, load all subdirectories found under the sub into a lower namespace
    :param hub: The redistributed pop central hub
    :param sub: The pop object that contains the loaded module data
    :param recurse: Recursively iterate over nested subs
    """
    if not sub._sub_virtual:
        return
    dirs = hub.pop.sub.get_dirs(sub)
    roots = {}
    for dir_ in dirs:
        for fn in os.listdir(dir_):
            if fn.startswith("_"):
                continue
            if fn == "contracts":
                continue
            if fn == "recursive_contracts":
                continue
            full = os.path.join(dir_, fn)
            if not os.path.isdir(full):
                continue
            if fn not in roots:
                roots[fn] = [full]
            else:
                roots[fn].append(full)
    for name, sub_dirs in roots.items():
        # Load er up!
        hub.pop.sub.add(
            subname=name,
            sub=sub,
            static=sub_dirs,
            virtual=sub._virtual,
            omit_start=sub._omit_start,
            omit_end=sub._omit_end,
            omit_func=sub._omit_func,
            omit_class=sub._omit_class,
            omit_vars=sub._omit_vars,
            mod_basename=sub._mod_basename,
            stop_on_failures=sub._stop_on_failures,
        )
        if recurse:
            if isinstance(getattr(sub, name), pop.hub.Sub):
                hub.pop.sub.load_subdirs(getattr(sub, name), recurse)


def reload(hub: pop.hub.Hub, subname: str):
    """
    Instruct the hub to reload the modules for the given sub. This does not call
    the init.new function or remove sub level variables. But it does re-read the
    directory list and re-initialize the loader causing all modules to be re-evaluated
    when started.
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    """
    if hasattr(hub, subname):
        sub = getattr(hub, subname)
        sub._prepare()
        return True
    else:
        return False


def extend(
    hub: pop.hub.Hub,
    subname: str,
    pypath: List[str] or str = None,
    static: List[str] or str = None,
    contracts_pypath: List[str] or str = None,
    contracts_static: List[str] or str = None,
) -> bool:
    """
    Extend the directory lookup for a given sub. Any of the directory lookup
    arguments can be passed.
    :param hub: The redistributed pop central hub
    :param subname: The name that the sub is going to take on the hub
        if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
    :param pypath: One or many python paths which will be imported
    :param static: Directories that can be explicitly passed
    :param contracts_pypath: Load additional contract paths
    :param contracts_static: Load additional contract paths from a specific directory
    """
    if not hasattr(hub, subname):
        return False
    sub = getattr(hub, subname)
    if pypath:
        sub._pypath.extend(pop.hub.ex_path(pypath))
    if static:
        sub._static.extend(pop.hub.ex_path(static))
    if contracts_pypath:
        sub._contracts_pypath.extend(pop.hub.ex_path(contracts_pypath))
    if contracts_static:
        sub._contracts_static.extend(pop.hub.ex_path(contracts_static))
    sub._prepare()
    return True


def dynamic(
    hub: pop.hub.Hub,
    resolver: Callable,
    context: Any = None,
    **kwargs,
):
    """
    Create dynamic (reverse) subs underneath the named sub.
    When a ReverseSub is called, it resolves it's current reference and calls the appropriate external function
    based on the current reference and context (as determined by the resolver).

    :param hub: The redistributed pop central hub
    :param resolver: A callable function that retrieves a target based on the current reference and context
    :param context: Resources that will be passed to the resolver function
    :param kwargs: kwargs to pass directly to hub.pop.sub.add
    """
    subname = kwargs.pop("subname", None)
    if kwargs.get("pypath") and not subname:
        pypath = pop.hub.ex_path(kwargs["pypath"])
        subname = pypath[0].split(".")[-1]
    elif kwargs.get("static"):
        subname = os.path.basename(kwargs["static"])
    if kwargs.get("dyne_name") and not subname:
        subname = kwargs["dyne_name"]

    root = kwargs.pop("sub", hub)

    if subname not in root._subs:
        # Only add reverse subs on top of conventional subs
        hub.pop.sub.add(sub=root, subname=subname, **kwargs)

    root._subs[subname]._reverse_sub = pop.hub.ReverseSub(
        hub=hub,
        resolver=resolver,
        context=context,
        subname=subname,
        root=root._subs[subname],
        sub_virtual=getattr(root._subs[subname], "_subvirt", True),
        **kwargs,
    )



"""
The main interface for management of the aio loop
"""
import asyncio
import signal
from typing import Any
from typing import AsyncGenerator
from typing import Callable
from typing import Coroutine
from typing import Iterable

import lazy_object_proxy

import pop.contract
import pop.hub

__virtualname__ = "loop"


def __init__(hub):
    hub.pop.loop.FUT_QUE = None
    hub.pop.loop.CURRENT_LOOP = None
    hub.pop.loop.EXECUTOR = None
    hub.pop.loop.POLICY = None
    hub.pop.sub.add(dyne_name="loop")
    hub.pop.Loop = lazy_object_proxy.Proxy(hub.loop.init.get)


def create(hub, loop_plugin: str = None):
    """
    Call the pop-loop dyne for creating the event loop.
    This is an idempotent operation.

    :param hub:
    :param loop_plugin: The plugin from pop-loop to use to initialize the loop
    """
    if not hub.pop.loop.CURRENT_LOOP:
        hub.loop.init.create(loop_plugin)


def call_soon(hub: "pop.hub.Hub", ref: str, *args, **kwargs):
    """
    Schedule a coroutine to be called when the loop has time. This needs
    to be called after the creation of the loop
    """
    coro = hub.pop.loop.wrap(hub[ref], *args, **kwargs)
    return hub.pop.loop.CURRENT_LOOP.call_soon(hub.pop.loop.unwrap, coro)


def ensure_future(hub: "pop.hub.Hub", ref: str, *args, **kwargs) -> asyncio.Future:
    """
    Schedule a coroutine to be called when the loop has time. This needs
    to be called after the creation of the loop. This function also uses
    the hold system to await the future when it is done making it easy
    to create a future that will be cleanly awaited in the background.
    """
    future = asyncio.ensure_future(hub[ref](*args, **kwargs))

    if hub.pop.loop.FUT_QUE is None:
        raise RuntimeError(
            "Run 'await hub.pop.loop.init_futures()' to create the futures queue in the proper loop"
        )

    def callback(fut):
        hub.pop.loop.FUT_QUE.put_nowait(fut)

    future.add_done_callback(callback)
    return future


async def init_futures(hub):
    if hub.pop.loop.FUT_QUE is None:
        hub.pop.loop.FUT_QUE = asyncio.Queue()


def start(
    hub: "pop.hub.Hub",
    *coros,
    hold: bool = False,
    sigint: Callable = None,
    sigterm: Callable = None,
    loop_plugin: asyncio.AbstractEventLoop = None,
) -> Any:
    """
    Start a loop that will run until complete
    """
    hub.pop.loop.create(loop_plugin=loop_plugin)
    loop: asyncio.AbstractEventLoop = hub.pop.loop.CURRENT_LOOP

    if sigint:
        s = signal.SIGINT
        loop.add_signal_handler(s, lambda s=s: loop.create_task(sigint(s)))
    if sigterm:
        s = signal.SIGTERM
        loop.add_signal_handler(s, lambda s=s: loop.create_task(sigterm(s)))
    if hold:
        coros = list(coros)
        coros.append(_holder(hub))

    try:
        # DO NOT CHANGE THIS CALL TO run_forever! If we do that then the tracebacks
        # do not get resolved.
        return loop.run_until_complete(asyncio.gather(*coros))
    except KeyboardInterrupt:
        print("Caught keyboard interrupt. Canceling...")
    finally:
        loop.close()


async def _holder(hub):
    """
    Just a sleeping while loop to hold the loop open while it runs until
    complete
    """
    if hub.pop.loop.FUT_QUE is None:
        hub.pop.loop.FUT_QUE = asyncio.Queue()

    while True:
        future = await hub.pop.loop.FUT_QUE.get()
        await future


async def await_futures(hub: "pop.hub.Hub"):
    """
    Scan over the futures that have completed and manually await them.
    This function is used to clean up futures when the loop is not opened
    up with hold=True so that ensured futures can be cleaned up on demand
    """
    if hub.pop.loop.FUT_QUE is None:
        hub.pop.loop.FUT_QUE = asyncio.Queue()

    while not hub.pop.loop.FUT_QUE.empty():
        future = await hub.pop.loop.FUT_QUE.get()
        await future


async def kill(hub: "pop.hub.Hub", wait: int or float = 0):
    """
    Close out the loop
    """
    await asyncio.sleep(wait)
    hub.pop.loop.CURRENT_LOOP.stop()
    while True:
        if hub.pop.loop.CURRENT_LOOP is not None:
            if not hub.pop.loop.CURRENT_LOOP.is_running():
                hub.pop.loop.CURRENT_LOOP.close()
                hub.pop.loop.CURRENT_LOOP = None
            else:
                await asyncio.sleep(1)


# Helpers for async operations


async def as_yielded(
    hub: "pop.hub.Hub", gens: Iterable[AsyncGenerator]
) -> AsyncGenerator:
    """
    Concurrently run multiple async generators and yield the next yielded
    value from the soonest yielded generator.

    async def many():
        for n in range(10):
            yield os.urandom(6).hex()

    async def run():
        gens = []
        for n in range(10):
            gens.append(many())
        async for y in as_yielded(gens):
            print(y)
    """
    fin = object()
    que = asyncio.Queue()
    to_clean = []

    async def _yield(agen: AsyncGenerator):
        async for comp in agen:
            await que.put(comp)

    async def _ensure(coroutines: Iterable[Coroutine]):
        for f in asyncio.as_completed(coroutines):
            await f

    async def _set_done():
        await que.put(fin)

    def _done(future: asyncio.Future):
        to_clean.append(asyncio.ensure_future(_set_done()))

    coros = []
    for gen in gens:
        coros.append(_yield(gen))

    fut = asyncio.ensure_future(_ensure(coros))
    fut.add_done_callback(_done)

    while True:
        ret = await que.get()
        if ret is fin:
            break
        yield ret
    for c in to_clean:
        await c


async def sleep(hub, delay: float, *args, **kwargs):
    await asyncio.sleep(delay, *args, **kwargs)


async def wrap(
    hub: "pop.hub.Hub",
    synchronous_function: Callable,
    *args,
    **kwargs,
):
    """
    Run a synchronous function asynchronously

    :param hub:
    :param synchronous_function: The function to wrap
    """
    return await hub.pop.loop.CURRENT_LOOP.run_in_executor(
        executor=hub.pop.loop.EXECUTOR,
        func=pop.contract.Partial(synchronous_function, *args, **kwargs),
    )


async def unwrap(hub, function_ret: Coroutine or Any):
    """
    Take the return of a function, if it is awaitable, await it.
    Return the result

    :param hub:
    :param function_ret: The return from a possibly asynchronous function
    """
    while asyncio.iscoroutine(function_ret):
        function_ret = await function_ret
    return function_ret



"""
Convenience wrappers to make using the conf system as easy and seamless as possible
"""
import warnings
from typing import Any
from typing import Dict
from typing import List

try:
    import pop_conf.conf  # noqa

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def __init__(hub):
    hub.pop.sub.add("pop_conf.conf")


def integrate(
    hub: "pop.hub.Hub",
    imports: List[str] or str,
    override: Dict[str, Any] = None,
    cli: str = None,
    roots: bool = None,
    loader: str = "json",
    logs: bool = True,
):
    """
    Load the conf sub and run the integrate sequence.
    """
    warnings.warn("Use `hub.pop.config.load()` instead", DeprecationWarning)
    hub.conf.integrate.load(
        imports, override, cli=cli, roots=roots, loader=loader, logs=logs
    )



"""
Used to scan the given directories for loadable files
"""
import collections
import importlib.machinery
import os
from typing import Any
from typing import Dict
from typing import Iterable

PY_END = (".py", ".pyc", ".pyo")
PYEXT_END = tuple(importlib.machinery.EXTENSION_SUFFIXES)
CYTHON_END = (".pyx",)
SKIP_DIRNAMES = ("__pycache__",)


def scan(dirs: Iterable[str]) -> Dict[str, Dict[str, Any]]:
    """
    :param dirs: A list of locations to search for importables files
    :return A description of importable files
    """
    ret = collections.OrderedDict()
    ret["python"] = collections.OrderedDict()
    ret["cython"] = collections.OrderedDict()
    ret["ext"] = collections.OrderedDict()
    ret["imp"] = collections.OrderedDict()
    for dir_ in dirs:
        for fn_ in os.listdir(dir_):
            _apply_scan(ret, dir_, fn_)
    return ret


def _apply_scan(
    ret: Dict[str, Dict[str, Any]], dir_: str, fn_: str
) -> None or Dict[str, Dict[str, Any]]:
    """
    Convert the scan data into paths and refs
    :param ret: The result of a scan()
    :param dir_:
    :param fn_:
    """
    if fn_.startswith("_"):
        return
    if os.path.basename(dir_) in SKIP_DIRNAMES:
        return
    full = os.path.join(dir_, fn_)
    if "." not in full:
        return
    bname = full[: full.rindex(".")]
    if fn_.endswith(PY_END):
        if bname not in ret["python"]:
            ret["python"][bname] = {"path": full}
    if fn_.endswith(CYTHON_END):
        if bname not in ret["cython"]:
            ret["cython"][bname] = {"path": full}
    if fn_.endswith(PYEXT_END):
        if bname not in ret["ext"]:
            ret["ext"][bname] = {"path": full}



"""
Used to resolve resolutions to paths on the hub
"""
from typing import List

import pop.hub


def last(hub: "pop.hub.Hub", ref: str) -> "pop.hub.Sub":
    """
    Takes a string that references the desired ref and returns the last object
    called out in that ref
    """
    return hub.pop.ref.path(ref)[-1]


def path(hub: "pop.hub.Hub", ref: str) -> List["pop.hub.Sub"]:
    """
    Retuns a list of references up to the named ref
    """
    ret = [hub]
    if isinstance(ref, str):
        ref = ref.split(".")
    for chunk in ref:
        ret.append(getattr(ret[-1], chunk))
    return ret


def create(hub: "pop.hub.Hub", ref: str, obj: object):
    """
    Create an attribute at a given target using just a ref string and the
    object to be saved at said location. The desired location must already
    exist!

    :param hub: The redistributed pop central hub
    :param ref: The dot delimited string referencing the target location to
        create the given object on the hub
    :param obj: The object to store at the given reference point
    """
    if "." not in ref:
        setattr(hub, ref, obj)
        return
    comps = ref.split(".")
    sub_ref = ref[: ref.rindex(".")]
    var = comps[-1]
    top = hub.pop.ref.last(sub_ref)
    setattr(top, var, obj)



"""
Load the files detected from the scanner
"""
import asyncio
import importlib.machinery
import importlib.util
import inspect
import os
import sys
import traceback as stdlib_traceback
import types
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple

import pop.contract
import pop.exc


class LoadError(Exception):
    """
    Errors from the loader are contained herein
    """

    __slots__ = ("edict", "traceback")

    def __init__(self, msg, exception=None, traceback=None, verror=None):
        self.edict = {
            "msg": msg,
            "exception": exception,
            "verror": verror,
        }
        self.traceback = traceback

    def __call__(self):
        """
        Return the error cases
        """
        return self.edict

    def __getattr__(self, attr):
        """
        Return a lambda that returns the edict
        """
        return self.__calling_load_error__

    def __calling_load_error__(
        self, *args, **kwargs
    ):  # pylint: disable=unused-argument
        if self.edict["verror"]:
            error = "{0[msg]}: {0[verror]}".format(self())
            raise pop.exc.PopError(error)
        error = "{0[msg]}: {0[exception]!r}".format(self())
        if self.traceback:
            error += "\n" + self.traceback
        raise pop.exc.PopError(error)

    def __repr__(self):
        return f"<{self.__class__.__name__} edict={self.edict!r}>"


def load_mod(modname: str, form: str, path: str) -> "LoadedMod":
    """
    Load a single module
    :param modname: The name of the module to get from the loader
    :param form: The name of the loader module
    :param path: The package to use as the anchor point from which to resolve the
        relative import to an absolute import.
    """
    this = sys.modules[__name__]
    return getattr(this, form)(modname, path)


def _generate_module(name: str) -> types.ModuleType:
    """
    Generate a module at runtime and insert it in sys.modules
    :param name: The name of the module to create
    """
    if name in sys.modules:
        return sys.modules[name]

    code = "'''POP sub auto generated parent module for {}'''".format(
        name.split(".")[-1]
    )
    # Create a new module that is not entered into sys.modules
    module = types.ModuleType(name)
    exec(code, module.__dict__)  # pylint: disable=exec-used
    sys.modules[name] = module
    return module


def _populate_sys_modules(mod: str):
    """
    This is a hack to populate sys.modules with the modules that pop loads
    while making sure that parent modules have the attribute for the child
    modules.
    """
    mod_parts = mod.split(".")
    imp_mod = mod_parts.pop(0)
    gen_mod = _generate_module(imp_mod)
    while True:
        if not mod_parts:
            break
        part = mod_parts.pop(0)
        imp_mod += "." + part
        gen_child_mod = _generate_module(imp_mod)
        setattr(gen_mod, part, gen_child_mod)
        gen_mod = gen_child_mod


def ext(modname: str, path: str) -> "LoadedMod" or LoadError:
    """
    Attempt to load the named python modules
    :param modname: The name of the module to get from the loader
    :param path: The package to use as the anchor point from which to resolve the
        relative import to an absolute import.
    """
    modname = ".".join(modname.split(".")[:-1])
    if modname in sys.modules:
        return sys.modules[modname]
    _populate_sys_modules(modname)
    try:
        efl = importlib.machinery.ExtensionFileLoader(modname, path)
        mod = efl.create_module(importlib.util.find_spec(modname))
        return efl.exec_module(mod)
    except Exception as exc:  # pylint: disable=broad-except
        return LoadError(
            f"Failed to load python module {modname} at path {path}",
            exception=exc,
            traceback=stdlib_traceback.format_exc(),
        )


def python(modname: str, path: str) -> "LoadedMod" or LoadError:
    """
    Attempt to load the named python modules
    :param modname: The name of the module to get from the loader
    :param path: The package to use as the anchor point from which to resolve the
        relative import to an absolute import.
    """
    if modname in sys.modules:
        return sys.modules[modname]
    try:
        sfl = importlib.machinery.SourceFileLoader(modname, path)
        return sfl.load_module()
    except Exception as exc:  # pylint: disable=broad-except
        return LoadError(
            f"Failed to load python module {modname} at path {path}",
            exception=exc,
            traceback=stdlib_traceback.format_exc(),
        )


def _base_name(bname: str, mod: "LoadedMod") -> Tuple[str, str]:
    """
    Find the basename and alias of a loader module
    :param bname: The base name of the mod's path
    :param mod: A loader module or a LoadError if the module didn't load
    """
    base_name = os.path.basename(bname)
    if "." in base_name:
        base_name = base_name.split(".")[0]
    return base_name, getattr(mod, "__virtualname__", base_name)


def _load_virtual(
    hub: "pop.hub.Hub",
    virtual: bool,
    mod: "LoadedMod" or LoadError,
    bname: str,
    vtype: str,
) -> Dict[str, Any]:
    """
    Run the virtual function to name the module and check for all loader
    errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    :param vtype: The name of the virtual function to call on the module I.E. __virtual__ or __sub_virtual__
    """
    base_name, name = _base_name(bname, mod)

    if isinstance(mod, LoadError):
        # The mod is a LoadError instance.
        # Return the load error with name as the base_name because another
        # module is still allowed to load under the same __virtualname__
        # but also return the vname information
        return {"name": base_name, "vname": name, "error": mod}

    if not virtual:
        # __virtual__ is not to be processed. Return now!
        return {"name": base_name}

    if not hasattr(mod, vtype):
        # No __virtual__ processing is required.
        # Return the mod's name as the defined __virtualname__ if defined,
        # else, the base_name
        return {"name": name}

    try:
        vret = getattr(mod, vtype)(hub)
        # If the __virtual__ function was asynchronous then run it in an asyncio loop
        if asyncio.iscoroutine(vret):
            vret = asyncio.get_running_loop().run_until_complete(vret)
    except Exception as exc:  # pylint: disable=broad-except
        err = LoadError(
            f"Virtual threw exception in mod {bname}",
            exception=exc,
            traceback=stdlib_traceback.format_exc(),
        )
        # Return the load error with name as the base_name because another
        # module is still allowed to load under the same __virtualname__
        # but also return the vname information
        return {"name": base_name, "vname": name, "error": err}

    verror = vret
    if isinstance(vret, tuple):
        if len(vret) > 1:
            verror = vret[1]
        vret = vret[0]

    if vret is True:
        # No problems occurred, module is allowed to load
        # Return the mod's name as the defined __virtualname__ if defined,
        # else, the base_name
        return {"name": name}

    if vret is False:
        # __virtual__ explicitly disabled the loading of this module
        err = LoadError(f"Module {bname} returned virtual FALSE", verror=verror)
        # Return the load error with name as the base_name because another
        # module is still allowed to load under the same __virtualname__
        # but also return the vname information
        return {"name": base_name, "vname": name, "error": err}

    # Anything else besides True/False should be considered a LoadError
    err = LoadError(f"Module {bname} returned virtual error", verror=verror)
    # Return the load error with name as the base_name because another
    # module is still allowed to load under the same __virtualname__
    # but also return the vname information
    return {"name": base_name, "vname": name, "error": err}


def load_virtual(
    hub: "pop.hub.Hub", virtual: bool, mod: "LoadedMod" or LoadError, bname: str
) -> Dict[str, Any]:
    """
    Run the __virtual__ function to name the module and check for all loader errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    """
    return _load_virtual(hub, virtual, mod, bname, "__virtual__")


def load_sub_virtual(
    hub: "pop.hub.Hub", virtual: bool, mod: "LoadedMod" or LoadError, bname: str
) -> Dict[str, Any]:
    """
    Run the __sub_virtual__ function to name the module and check for all loader errors
    :param hub: The redistributed pop central hub
    :param virtual: Toggle whether or not to process __virtual__ functions
    :param mod: A loader module or a LoadError if the module didn't load
    :param bname: The base name of the mod's path
    """
    _, name = _base_name(bname, mod)
    if name != "init":
        return {"name": name}
    return _load_virtual(hub, virtual, mod, bname, "__sub_virtual__")


def mod_init(sub: "pop.hub.Sub", mod: "LoadedMod", mod_name: str):
    """
    Process module's __init__ function if defined
    :param sub: The pop object that contains the loaded module data
    :param mod: A loader modul
    :param mod_name: The name of the module to get from the loader
    """
    if "__init__" in dir(mod):
        init = pop.contract.Contracted(
            sub._hub,
            contracts=[],
            func=mod.__init__,
            ref=f"{sub._subname}.{mod_name}",
            name="__init__",
        )
        ret = init()
        # If the __init__ function was asynchronous then run it in an asyncio loop
        if asyncio.iscoroutine(ret):
            asyncio.get_running_loop().run_until_complete(ret)


def sub_alias(this_sub: "pop.hub.Sub", mod: "LoadedMod", mod_name: str):
    """
    Check the sub alias settings and apply the alias names locally so they can be gathered into the higher level object on the hub
    :param this_sub: The pop object that contains the loaded module data
    :param mod: A loader module
    :param mod_name: The name of the module to get from the loader
    """
    if mod_name == "init":
        alias = getattr(mod, "__sub_alias__", [])
        if alias:
            this_sub._alias = alias


def prep_loaded_mod(
    this_sub: "pop.hub.Sub",
    mod: "LoadedMod",
    mod_name: str,
    contracts: "List[pop.contract.Wrapper]",
    recursive_contracts: "List[pop.contract.Wrapper]",
) -> "LoadedMod":
    """
    Read the attributes of a python module and create a LoadedMod, which resolves
    aliases and omits objects that should not be exposed.
    :param this_sub: The pop object that contains the loaded module data
    :param mod: A loader module
    :param mod_name: The name of the module to get from the loader
    :param contracts: Contracts functions to add to the sub
    :param recursive_contracts:
    """
    lmod = this_sub._loaded.get(mod_name, LoadedMod(mod_name))
    ref = f"{this_sub._subname}.{mod_name}"  # getattr(hub, ref) should resolve to this module

    # Recursively get the full reference to this mod
    root = this_sub._root
    while hasattr(root, "_subname"):
        ref = f"{root._subname}.{ref}"
        root = root._root

    sub_alias(this_sub, mod, mod_name)

    __func_alias__: Callable or Dict[str, str or Callable] = getattr(
        mod, "__func_alias__", {}
    )
    # If __func_alias__ is a function, run it and replace the function with it's returned dictionary
    if inspect.isfunction(__func_alias__):
        __func_alias__ = __func_alias__(this_sub._hub)

    # Get all the attributes from the loaded mod
    for attr in getattr(mod, "__load__", dir(mod)):
        # For an attribute in the module, get it's function alias
        func_alias: str = __func_alias__.get(attr, attr)
        func = getattr(mod, attr)
        name = func_alias
        if not this_sub._omit_vars:
            if (
                not inspect.isfunction(func)
                and not inspect.isclass(func)
                and type(func).__name__ != "cython_function_or_method"
            ):
                lmod._vars[name] = func
                lmod._attrs[name] = func
                continue
        if attr.startswith(this_sub._omit_start):
            continue
        if attr.endswith(this_sub._omit_end):
            continue
        if (
            inspect.isfunction(func)
            or inspect.isbuiltin(func)
            or type(func).__name__ == "cython_function_or_method"
        ):
            obj = pop.contract.create_contracted(
                this_sub._hub,
                contracts + recursive_contracts,
                func,
                ref,
                name,
            )
            if not this_sub._omit_func:
                if this_sub._pypath and not func.__module__.startswith(mod.__name__):
                    # We're only interested in functions defined in this module, not
                    # imported functions
                    continue
                lmod._funcs[name] = obj
                lmod._attrs[name] = obj
        else:
            klass = func
            if not this_sub._omit_class and inspect.isclass(klass):
                # We're only interested in classes defined in this module, not
                # imported classes
                if not klass.__module__.startswith(mod.__name__):
                    continue
                lmod._classes[name] = klass
                lmod._attrs[name] = klass

    # Add external aliased functions to the loaded mod
    for attr, func_alias in __func_alias__.items():
        if isinstance(func_alias, Callable):
            obj = pop.contract.create_contracted(
                this_sub._hub,
                contracts + recursive_contracts,
                func_alias,
                ref,
                attr,
                implicit_hub=False,
            )
            lmod._funcs[attr] = obj
            lmod._attrs[attr] = obj

    return lmod


class LoadedMod(types.ModuleType):
    """
    The LoadedMod class allows for the module loaded onto the sub to return
    custom sequencing, for instance it can be iterated over to return all
    functions
    """

    def __init__(self, name: str):
        super().__init__(name)
        self._vars = {}
        self._funcs = {}
        self._classes = {}
        self._attrs = {}

    def __getattr__(self, item: str):
        if item in self._attrs:
            return self._attrs[item]
        raise AttributeError(item)

    def __getitem__(self, item: str):
        return getattr(self, item)

    def __iter__(self):
        keys = sorted(self._funcs)
        ret = []
        for key in keys:
            ret.append(self._funcs[key])
        return iter(ret)

    def __dir__(self):
        # TODO: This should return finite set attrs as well as dunder attrs
        ret = list(self._attrs.keys())
        ret.extend(["__name__", "_vars", "_funcs", "_classes", "_attrs"])
        return ret



from typing import List

import pop.hub


def __virtual__(hub):
    hub.pop.sub.add(dyne_name="config")
    return hasattr(hub.config, "integrate"), "pop-config is not installed"


def load(
    hub: "pop.hub.Hub",
    sources: List[str],
    cli: str = None,
    dyne_name: str = None,
    loader: str = "yaml",
    parse_cli: bool = True,
    logs: bool = True,
):
    """
    Use the pop-config system to load up a fresh configuration for this project
    from the included conf.py file.
    """
    hub.config.integrate.load(sources, cli, dyne_name, loader, parse_cli, logs)



"""
Tools to work with dicts
"""
import warnings

import dict_tools.data
import dict_tools.update

import pop.hub


def traverse(hub: "pop.hub.Hub", data, key, default=None, delimiter=":"):
    """
    Traverse a dict or list using a colon-delimited (or otherwise delimited,
    using the 'delimiter' param) target string. the target 'foo:bar:0' will
    return data['foo']['bar'][0] if this value exists, and will otherwise
    return the dict in the default argument.
    function will automatically determine the target type.
    the target 'foo:bar:0' will return data['foo']['bar'][0] if data like
    {'foo':{'bar':['baz']}} , if data like {'foo':{'bar':{'0':'baz'}}}
    then return data['foo']['bar']['0']
    """
    warnings.warn(
        "Use `dict_tools.data.traverse_dict_and_list()` instead", DeprecationWarning
    )
    return dict_tools.data.traverse_dict_and_list(data, key, default.delimiter)


def update(hub: "pop.hub.Hub", dest, upd, recursive_update=True, merge_lists=True):
    """
    Recursive version of the default dict.update

    Merges upd recursively into dest

    If recursive_update=false, will use the classic dict.update, or fall back
    on a manual merge (helpful for non-dict types like functionwrapper)

    If merge_lists=true, will aggregate list object types instead of replace.
    the list in ``upd`` is added to the list in ``dest``, so the resulting list
    is ``dest[key] + upd[key]``. This behavior is only activated when
    recursive_update=True. By default merge_lists=False.
    """
    warnings.warn(
        "Use `dict_tools.update.update()` instead", DeprecationWarning, stacklevel=1
    )
    return dict_tools.update.update(dest, upd, recursive_update, merge_lists)



"""
The input module is used to translate typical input strings into the
ref/args/kwargs used by pop when forwarding data into functions.
"""
import re
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import yaml

import pop.hub

KWARG_REGEX = re.compile(r"^([^\d\W][\w.-]*)=(?!=)(.*)$", re.UNICODE)


def parse(
    hub: "pop.hub.Hub",
    args: List[Any],
    condition: bool = True,
    no_parse: Tuple[str] = None,
) -> Tuple[List[Any], Dict[str, Any]]:
    """
    Parse out the args and kwargs from a list of input values. Optionally,
    return the args and kwargs without passing them to condition_input().
    Don't pull args with key=val apart if it has a newline in it.
    """
    if no_parse is None:
        no_parse = ()
    _args = []
    _kwargs = {}
    for arg in args:
        if isinstance(arg, str):
            if "=" in arg:
                arg_name, arg_value = _parse_kwarg(arg)
                if arg_name:
                    _kwargs[arg_name] = (
                        _yamlify_arg(arg_value)
                        if arg_name not in no_parse
                        else arg_value
                    )
            else:
                _args.append(_yamlify_arg(arg))
        elif isinstance(arg, dict):
            # Yes, we're popping this key off and adding it back if
            # condition_input is called below, but this is the only way to
            # gracefully handle both CLI and API input.
            if arg.pop("__kwarg__", False) is True:
                _kwargs.update(arg)
            else:
                _args.append(arg)
        else:
            _args.append(arg)
    if condition:
        return _condition_input(_args, _kwargs)
    return _args, _kwargs


def _yamlify_arg(arg: Any) -> Any:
    """
    yaml.safe_load the arg
    """
    if not isinstance(arg, str):
        return arg

    if arg.strip() == "":
        # Because YAML loads empty (or all whitespace) strings as None, we
        # return the original string
        # >>> import yaml
        # >>> yaml.load('') is None
        # True
        # >>> yaml.load('      ') is None
        # True
        return arg

    elif "_" in arg and all([x in "0123456789_" for x in arg.strip()]):
        # When the stripped string includes just digits and underscores, the
        # underscores are ignored and the digits are combined together and
        # loaded as an int. We don't want that, so return the original value.
        return arg

    try:
        original_arg = arg
        if "#" in arg:
            # Only yamlify if it parses into a non-string type, to prevent
            # loss of content due to # as comment character
            parsed_arg = yaml.safe_load(arg)
            if isinstance(parsed_arg, str) or parsed_arg is None:
                return arg
            return parsed_arg
        if arg == "None":
            arg = None
        else:
            arg = yaml.safe_load(arg)

        if isinstance(arg, dict):
            # dicts must be wrapped in curly braces
            if isinstance(original_arg, str) and not original_arg.startswith("{"):
                return original_arg
            else:
                return arg

        elif isinstance(arg, list):
            # lists must be wrapped in brackets
            if isinstance(original_arg, str) and not original_arg.startswith("["):
                return original_arg
            else:
                return arg

        elif arg is None or isinstance(arg, (list, float, int, str)):
            # yaml.safe_load will load '|' as '', don't let it do that.
            if arg == "" and original_arg in ("|",):
                return original_arg
            # yaml.safe_load will treat '#' as a comment, so a value of '#'
            # will become None. Keep this value from being stomped as well.
            elif arg is None and original_arg.strip().startswith("#"):
                return original_arg
            else:
                return arg
        else:
            # we don't support this type
            return original_arg
    except Exception:
        # In case anything goes wrong...
        return original_arg


def _parse_kwarg(string_: str) -> Tuple:
    """
    Parses the string and looks for the following kwarg format:
    "{argument name}={argument value}"
    For example: "my_message=Hello world"
    Returns the kwarg name and value, or (None, None) if the regex was not
    matched.
    """
    try:
        return KWARG_REGEX.match(string_).groups()
    except AttributeError:
        return None, None


def _condition_input(args: List[Any], kwargs: Dict[str, Any]) -> List[str]:
    """
    Return a single arg structure for the publisher to safely use
    """
    ret = []
    for arg in args:
        if isinstance(arg, int):
            ret.append(str(arg))
        else:
            ret.append(arg)
    if isinstance(kwargs, dict) and kwargs:
        kw_ = {"__kwarg__": True}
        for key, val in kwargs.items():
            kw_[key] = val
        return ret + [kw_]
    return ret



"""
Find directories
"""
import importlib
import os
import sys
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List

import dict_tools.update


def dir_list(
    subname: str, p_name: str, pypath: List[str] = None, static: List[str] = None
) -> List[str]:
    """
    Return the directories to look for modules in, pypath specifies files
    relative to an installed python package, static is for static dirs
    :param subname: ignored
    :param p_name: ignored
    :param pypath: One or many python paths which will be imported
    :param static: Directories that can be explicitly passed
    """
    ret = []
    for path in pypath:
        mod = importlib.import_module(path)
        for m_path in mod.__path__:
            # If we are inside of an executable the path will be different
            ret.append(m_path)
    ret.extend(static)
    return ret


def inline_dirs(dirs: Iterable[str], subdir: str) -> List[str]:
    """
    Look for the named subdir in the list of dirs
    :param dirs: The names of configured dynamic dirs
    :param subdir: The name of the subdir to check for in the list of dynamic dirs
    :return An extended list of dirs that includes the found subdirs
    """
    ret = []
    for dir_ in dirs:
        check = os.path.join(dir_, subdir)
        if os.path.isdir(check):
            ret.append(check)
    return ret


def dynamic_dirs() -> Dict[str, Any]:
    """
    Iterate over the available python package imports and look for configured
    dynamic dirs
    """
    dirs = []
    ret = {}
    for dir_ in sys.path:
        if not os.path.isdir(dir_):
            continue
        for sub in os.listdir(dir_):
            full = os.path.join(dir_, sub)
            if full.endswith(".egg-link"):
                with open(full) as rfh:
                    dirs.append(rfh.read().strip())
            if os.path.isdir(full):
                dirs.append(full)
    for dir_ in dirs:
        conf = os.path.join(dir_, "conf.py")
        context = {}
        if not os.path.isfile(conf):
            continue
        try:
            with open(conf) as f:
                code = f.read()
                if "DYNE" in code:
                    exec(code, context)
                else:
                    continue
        except Exception:
            continue
        if "DYNE" in context:
            if not isinstance(context["DYNE"], dict):
                continue
            for name, paths in context["DYNE"].items():
                if not isinstance(paths, list):
                    continue
                if name not in ret:
                    ret[name] = {
                        "paths": [],
                        "CONFIG": {},
                        "CLI_CONFIG": {},
                        "SUBCOMMANDS": {},
                    }
                if "CONFIG" in context:
                    dict_tools.update.update(ret[name]["CONFIG"], context["CONFIG"])
                if "CLI_CONFIG" in context:
                    dict_tools.update.update(
                        ret[name]["CLI_CONFIG"], context["CLI_CONFIG"]
                    )
                if "SUBCOMMANDS" in context:
                    dict_tools.update.update(
                        ret[name]["SUBCOMMANDS"], context["SUBCOMMANDS"]
                    )
                for path in paths:
                    ref = os.path.join(dir_, path.replace(".", os.sep))
                    if dir_.endswith(name):
                        ret[name]["paths"].insert(0, ref)
                    else:
                        ret[name]["paths"].append(ref)
    for name in ret:
        if not ret[name]:
            continue
        first = ret[name]["paths"].pop(0)
        ret[name]["paths"] = sorted(ret[name]["paths"])
        ret[name]["paths"].insert(0, first)
    return ret



import copy
import importlib.machinery
import inspect
import logging
import os
import secrets
import sys
from types import ModuleType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple

import pop.contract
import pop.dirs
import pop.exc
import pop.loader
import pop.scanner
import pop.verify

try:
    import pytest_pop.mods  # noqa

    HAS_TESTING_LIBS = True
except ImportError:
    HAS_TESTING_LIBS = False

EXT_SUFFIXES = tuple(importlib.machinery.EXTENSION_SUFFIXES)
log = logging.getLogger(__name__)


def ex_path(path: str) -> List[str]:
    """
    Take a path that is sent to the Sub and expand it if it is a string or not
    """
    if path is None:
        return []
    elif isinstance(path, str):
        return path.split(",")
    elif isinstance(path, list):
        return path
    return []


class Hub:
    """
    The redistributed pop central hub. All components of the system are
    rooted to the Hub.
    """

    def __init__(self):
        self._subs = {}
        self._sub_alias = {}
        self._dynamic = {}
        self._dscan = False
        # Add the pop sub to the hub, this should always use pypath and
        # Should never be made dynamic. This is a core system sub and should
        # NOT be app-merged
        pypaths = ["pop.mods.pop"]
        if HAS_TESTING_LIBS:
            pypaths.append("pytest_pop.mods")
        self._subs["pop"] = Sub(self, "pop", pypath=pypaths)
        self._reverse_sub = None
        self._iter_ind = 0
        # Set up the conf OPT structure so it is always available
        self.OPT = {}

    @property
    def _iter_subs(self):
        return sorted(self._subs.keys())

    def __getstate__(self) -> Dict:
        return dict(
            reverse_sub=self._reverse_sub,
            subs=self._subs,
            OPT=copy.copy(self.OPT),
        )

    def __setstate__(self, state: Dict):
        self.__init__()
        subs = state["subs"]
        opt = state["OPT"]
        reverse_sub = state["reverse_sub"]

        self.OPT = self.pop.data.imap(opt)
        self._restore_subs(subs, self)
        if reverse_sub:
            self.pop.sub.dynamic(
                sub=self,
                resolver=reverse_sub._resolver,
                context=reverse_sub._context,
                **reverse_sub._restoreable_state,
            )

    def _restore_subs(self, subs: Dict[str, "Sub"], root):
        """
        Recursively restore all pickled subs that were added to this hub.
        Make sure they are prepared properly with inits/virtuals/loadedmods
        """
        for name, sub in subs.items():
            if sub._reverse_sub:
                self.pop.sub.dynamic(
                    resolver=sub._reverse_sub._resolver,
                    context=sub._reverse_sub._context,
                    sub=root,
                    **sub._reverse_sub._restoreable_state,
                )
            else:
                self.pop.sub.add(sub=root, load_all=True, **sub._restoreable_state)
            self._restore_subs(sub._subs, root[name])

    def __iter__(self) -> Iterator["Sub"]:
        def iterator(subs: Dict[str, Sub]):
            for sub in self._iter_subs:
                yield subs[sub]

        return iterator(self._subs)

    def _resolve_this(self, levels: int) -> "Hub":
        """
        This function allows for hub to pop introspective calls.
        This should only ever be called from within a hub module, otherwise
        it should stack trace, or return heaven knows what...
        :param levels: The number of frames to search for a hub reference
        """
        if hasattr(
            sys, "_getframe"
        ):  # implementation detail of CPython, speeds up things by 100x.
            desired_frame = sys._getframe(3)
            contracted = desired_frame.f_locals["self"]
        else:
            call_frame = inspect.stack(0)[3]
            contracted = call_frame[0].f_locals["self"]
        ref = contracted.ref.split(".")

        # (0=module, 1=module's parent etc.)
        level_offset = levels - 1
        traversed = self
        for i in range(len(ref) - level_offset):
            traversed = getattr(traversed, ref[i])
        return traversed

    def _remove_subsystem(self, subname: str) -> bool:
        """
        Remove the named subsystem
        :param subname: The name of a subsystem to remove
        :return True if the subsystem was successfully removed, else False
        """
        if subname in self._subs:
            # Remove the subsystem
            self._subs.pop(subname)
            # reset the iterator
            self._iter_ind = 0
            return True
        return False

    def _scan_dynamic(self):
        """
        Refresh the dynamic roots data used for loading app merge module roots
        """
        self._dynamic = pop.dirs.dynamic_dirs()
        self._dscan = True

    def __getattr__(self, item: str):
        if item.startswith("_"):
            if item == item[0] * len(item):
                return self._resolve_this(len(item))
            else:
                return self.__getattribute__(item)
        if "." in item:
            return self.pop.ref.last(item)
        if item in self._subs:
            return self._subs[item]
        elif item in self._sub_alias:
            resolved = self._sub_alias[item]
            if resolved in self._subs:
                return self._subs[resolved]
        return self.__getattribute__(item)

    def __getitem__(self, item: str):
        return getattr(self, item)


class Sub:
    """
    The pop object that contains the loaded module data
    """

    def __init__(
        self,
        hub: Hub,
        subname: str,
        root: Hub or "Sub" = None,
        pypath: List[str] or str = None,
        static: List[str] or str = None,
        contracts_pypath: List[str] or str = None,
        contracts_static: List[str] or str = None,
        default_contracts: List[str] or str = None,
        virtual: bool = True,
        dyne_name: str = None,
        omit_start: Tuple[str] = ("_",),
        omit_end: Tuple[str] = (),
        omit_func: bool = False,
        omit_class: bool = False,
        omit_vars: bool = False,
        mod_basename: str = "",
        stop_on_failures: bool = False,
        init: bool = True,
        is_contract: bool = False,
        sub_virtual: bool = True,
        recursive_contracts_static=None,
        default_recursive_contracts=None,
    ):
        """
        :param hub: The redistributed pop central hub
        :param subname: The name that the sub is going to take on the hub
            if nothing else is passed, it is used as the pypath (TODO make it the dyne_name not the pypath)
        :param pypath: One or many python paths which will be imported
        :param static: Directories that can be explicitly passed
        :param contracts_pypath: Load additional contract paths
        :param contracts_static: Load additional contract paths from a specific directory
        :param default_contracts: Specifies that a specific contract plugin will be applied as a default to all plugins
        :param virtual: Toggle whether or not to process __virtual__ functions
        :param dyne_name: The dynamic name to use to look up paths to find plugins -- linked to conf.py
        :param omit_start: Allows you to pass in a tuple of characters that would omit the loading of any object
            I.E. Any function starting with an underscore will not be loaded onto a plugin
            (You should probably never change this)
        :param omit_end:Allows you to pass in a tuple of characters that would omit the loading of an object
            (You should probably never change this)
        :param omit_func: bool: Don't load any functions
        :param omit_class: bool: Don't load any classes
        :param omit_vars: bool: Don't load any vars
        :param mod_basename: str: Manipulate the location in sys.modules that the plugin will be loaded to.
            Allow plugins to be loaded into a separate namespace.
        :param stop_on_failures: If any module fails to load for any reason, stacktrace and do not continue loading this sub
        :param init: bool: determine whether or not we process __init__ functions
        :param is_contract: Specify whether or not this sub is a contract
        :param sub_virtual: bool: Recursively ignore this sub and it's subs
        """
        self._iter_ind = 0
        self._hub = hub
        self._root = root or hub
        self._subs = {}
        self._alias = []
        self._sub_alias = {}
        self._subname = subname
        self._pypath = ex_path(pypath)
        self._static = ex_path(static)
        self._contracts_pypath = ex_path(contracts_pypath)
        self._contracts_static = ex_path(contracts_static)
        self._recursive_contracts_static = ex_path(recursive_contracts_static)
        if isinstance(default_contracts, str):
            default_contracts = [default_contracts]
        if isinstance(default_recursive_contracts, str):
            default_recursive_contracts = [default_recursive_contracts]
        self._reverse_sub = None
        self._default_recursive_contracts = default_recursive_contracts or []
        self._default_contracts = default_contracts or ()
        self._dyne_name = dyne_name
        self._virtual = virtual
        self._omit_start = omit_start
        self._sub_virtual = sub_virtual
        self._omit_end = omit_end
        self._omit_func = omit_func
        self._omit_class = omit_class
        self._omit_vars = omit_vars
        self._mod_basename = mod_basename
        self._stop_on_failures = stop_on_failures
        self._is_contract = is_contract
        self._process_init = init
        self._prepare()

    def _prepare(self):
        self._dirs = pop.dirs.dir_list(
            self._subname,
            "mods",
            self._pypath,
            self._static,
        )
        if self._dyne_name:
            self._load_dyne()
        self._contract_dirs = pop.dirs.dir_list(
            self._subname,
            "contracts",
            self._contracts_pypath,
            self._contracts_static,
        )
        self._contract_dirs.extend(pop.dirs.inline_dirs(self._dirs, "contracts"))
        self._recursive_contract_dirs = pop.dirs.dir_list(
            self._subname,
            "recursive_contracts",
            [],
            self._recursive_contracts_static,
        )
        self._recursive_contract_dirs.extend(
            pop.dirs.inline_dirs(self._dirs, "recursive_contracts")
        )

        if self._contract_dirs:
            self._contracts = Sub(
                hub=self._hub,
                subname=f"{self._subname}.contracts",
                static=self._contract_dirs,
                is_contract=True,
            )
        else:
            self._contracts = None

        if self._recursive_contract_dirs:
            self._recursive_contracts = Sub(
                hub=self._hub,
                subname=f"{self._subname}.recursive_contracts",
                static=self._recursive_contract_dirs,
                is_contract=True,
            )
        else:
            self._recursive_contracts = getattr(
                self._root, "_recursive_contracts", None
            )
        self._name_root = self._load_name_root()
        self._scan = pop.scanner.scan(self._dirs)
        self._loaded = {}
        self._vmap = {}
        self._load_errors = {}
        self._loaded_all = False

    def _load_dyne(self):
        """
        Load up the dynamic dirs for this sub
        """
        if not self._hub._dscan:
            self._hub._scan_dynamic()
        for path in self._hub._dynamic.get(self._dyne_name, {}).get("paths", []):
            self._dirs.append(path)

    def _load_name_root(self):
        """
        Generate the root of the name to be used to apply to the loaded modules
        """
        if self._pypath:
            return self._pypath[0]
        elif self._dirs:
            return secrets.token_hex()

    def __getstate__(self):
        return dict(
            pypath=self._pypath,
            subname=self._subname,
            static=self._static,
            contracts_pypath=self._contracts_pypath,
            contracts_static=self._contracts_static,
            default_contracts=self._default_contracts,
            virtual=self._virtual,
            dyne_name=self._dyne_name,
            omit_start=self._omit_start,
            omit_end=self._omit_end,
            omit_func=self._omit_func,
            omit_class=self._omit_class,
            omit_vars=self._omit_vars,
            mod_basename=self._mod_basename,
            stop_on_failures=self._stop_on_failures,
            init=self._process_init,
            recursive_contracts_static=self._recursive_contracts_static,
            default_recursive_contracts=self._default_recursive_contracts,
            _subs=self._subs,
            _reverse_sub=self._reverse_sub,
        )

    def __setstate__(self, state: Dict):
        """
        Only a hub can truly restore a sub because of their interdependency
        """
        self._subs = state.pop("_subs")
        self._reverse_sub = state.pop("_reverse_sub")
        # This will have all the arguments needed to recreate the Sub with pop.sub.add
        self._restoreable_state = state

    def __getattr__(self, item: str):
        """
        If the item should be loaded, load it, else serve it
        """
        if item.startswith("_"):
            return self.__getattribute__(item)
        elif "." in item:
            # Get the attribute from one sub higher up until the hub calls "ref.last"
            return getattr(self._root, f"{self._subname}.{item}")
        elif item in self._loaded:
            ret = self._loaded[item]
            # If this previously errored on load, try it again,
            # it might be ready to load now
            if isinstance(ret, pop.loader.LoadError):
                ret = self._find_mod(item)
                if isinstance(ret, pop.loader.LoadError):
                    # If this is still a LoadError, process it
                    self._process_load_error(ret)
            return ret
        elif item in self._subs:
            return self._subs[item]
        elif item in self._sub_alias:
            resolved = self._sub_alias[item]
            if resolved in self._subs:
                return self._subs[resolved]

        mod = self._find_mod(item)
        if mod is None:
            if self._reverse_sub:
                # This sub contains a reverse sub that could contain the result
                return self._reverse_sub[item]
            raise AttributeError(f"'{self._subname}' has no attribute '{item}'")
        return mod

    def __getitem__(self, item: str):
        return getattr(self, item)

    def __contains__(self, item: str):
        try:
            ret = getattr(self, item, None)
            if isinstance(ret, ReverseSub):
                return item in self._reverse_sub
            return ret is not None
        except pop.exc.PopLookupError:
            return False

    def __iter__(self) -> Iterator["Sub"]:
        self._load_all()

        def iterate(loaded):
            for mod_name in sorted(loaded.keys()):
                yield loaded[mod_name]

        return iterate(self._loaded)

    def __next__(self) -> "Sub":
        self._load_all()
        if self._iter_ind == len(self._iter_keys):
            self._iter_ind = 0
            raise StopIteration
        self._iter_ind += 1
        return self._loaded[self._iter_keys[self._iter_ind - 1]]

    def _sub_init(self):
        """
        Run load init.py for the sub, running '__init__' function if present
        """
        self._find_mod("init", match_only=True)

    def _process_load_error(
        self, mod: ModuleType, skip_full_stop: bool = False
    ) -> bool:
        if not isinstance(mod, pop.loader.LoadError):
            # This is not a LoadError, return now!
            return False

        if mod.edict["verror"]:
            error = "{0[msg]}: {0[verror]}".format(mod())
            if skip_full_stop is False and self._stop_on_failures is True:
                raise pop.exc.PopError(error)
            log.log(level=5, msg=error)
            return False
        error = "{0[msg]}: {0[exception]!r}".format(mod())
        if mod.traceback:
            error += "\n" + mod.traceback
        if skip_full_stop is False and self._stop_on_failures is True:
            raise pop.exc.PopError(error)
        if mod.traceback:
            log.warning(error)
        else:
            log.info(error)
        return True

    def _find_mod(self, item: str, match_only: bool = False) -> Dict:
        """
        Find the module named item
        :param item: The module to search for (then load) from any scanned interface
        :param match_only: return the loaded module
        :return a loaded mod_dict
        """
        for iface in self._scan:
            for bname in self._scan[iface]:
                if os.path.basename(bname) == item:
                    self._load_item(iface, bname)
            if item in self._loaded:
                return self._loaded[item]
        if not match_only:
            for iface in self._scan:
                for bname in self._scan[iface]:
                    if self._scan[iface][bname].get("loaded"):
                        continue
                    self._load_item(iface, bname)
                    if item in self._loaded:
                        return self._loaded[item]
        # Let's see if the module being lookup is in the load errors dictionary
        if item in self._load_errors:
            # Return the LoadError
            return self._load_errors[item]

    def _load_item(self, iface: str, bname: str):
        """
        Load the named basename
        :param iface: A scanned directory type
        :param bname: The base name of the python path of a module
        """
        if iface not in self._scan:
            raise pop.exc.PopLoadError(f"Bad call to load item, no iface {iface}")
        if bname not in self._scan[iface]:
            raise pop.exc.PopLoadError(
                f"Bad call to load item, no bname {bname} in iface {iface}"
            )
        # The mname is the name to give the module in python's sys.modules
        # This name must be unique for every loaded module, so we use the full
        # module path sans the file extention
        mname = self._scan[iface][bname]["path"].replace(os.sep, ".")
        mname = mname[mname.index(".") + 1 : mname.rindex(".")].strip(".")
        mod = pop.loader.load_mod(
            mname,
            iface,
            self._scan[iface][bname]["path"],
        )
        if self._process_load_error(mod):
            self._load_errors[os.path.basename(bname)] = mod
            return
        self._prep_mod(mod, iface, bname)

    def _process_vret(self, vret: Dict[str, Any]) -> bool:
        """
        :param vret: The return from a __virtual__ or __sub_virtual__ function
        :return: True if there was an error, else false
        """
        if "error" in vret:
            # Virtual Errors should not full stop pop
            self._process_load_error(vret["error"], skip_full_stop=True)
            # Store the LoadError under the __virtualname__ if defined
            self._load_errors[vret["vname"]] = vret["error"]
            return True
        else:
            return False

    def _prep_mod(self, mod: ModuleType, iface: str, bname: str):
        """
        Prepare the module!
        :param mod: A python module containing data
        :param iface: A scanned directory type
        :param bname: The base name of the python path of a module
        """
        if not self._sub_virtual:
            return
        else:
            vret = pop.loader.load_sub_virtual(self._hub, self._virtual, mod, bname)
            if self._process_vret(vret):
                self._sub_virtual = False
                return
        vret = pop.loader.load_virtual(self._hub, self._virtual, mod, bname)
        if self._process_vret(vret):
            return

        contracts = pop.contract.load_contract(
            self._contracts, self._default_contracts, mod, vret["name"]
        )
        recursive_contracts = set(
            pop.contract.load_contract(
                self._recursive_contracts,
                self._default_recursive_contracts,
                mod,
                vret["name"],
            )
        )
        if getattr(self._root, "_recursive_contracts", None):
            recursive_contracts.update(
                pop.contract.load_contract(
                    self._root._recursive_contracts,
                    self._root._default_recursive_contracts,
                    mod,
                    vret["name"],
                )
            )
        recursive_contracts = list(recursive_contracts)
        name = vret["name"]
        if name.endswith(EXT_SUFFIXES):
            for ext in EXT_SUFFIXES:
                if name.endswith(ext):
                    name = name.split(ext)[0]
                    break
        mod_dict = pop.loader.prep_loaded_mod(
            self, mod, name, contracts, recursive_contracts
        )
        if name != "init":
            pop.verify.contract(self._hub, contracts + recursive_contracts, mod_dict)
        self._loaded[name] = mod_dict
        self._vmap[mod.__file__] = name
        # Let's mark the module as loaded
        self._scan[iface][bname]["loaded"] = True
        if self._process_init:
            # Now that the module has been added to the sub, call mod_init
            pop.loader.mod_init(self, mod, name)

    def _load_all(self):
        """
        Load all modules found during the scan.

        .. attention:: This completely disables the lazy loader behavior of pop
        """
        if self._loaded_all is True:
            return
        for iface in self._scan:
            for bname in self._scan[iface]:
                if self._scan[iface][bname].get("loaded"):
                    continue
                self._load_item(iface, bname)
        self._loaded_all = True


class ReverseSub(Sub):
    """
    A sub that discovers attributes available to it dynamically through a resolver
    """

    def __init__(
        self,
        hub: Hub,
        subname: str,
        resolver: Callable,
        context: Any = None,
        refs: Tuple[str, ...] = None,
        **kwargs,
    ):
        """
        :param subname: The name that the sub is going to take on the hub
        :param resolver: A callable function that retrieves a target based on the current reference and context
        :param context: Resources that will be passed to the resolver function
        :param kwargs: Any kwargs that should be passed to the underlying sub
        """
        Sub.__init__(self, subname=subname, hub=hub, **kwargs)
        self._resolver = resolver
        self._context = context
        # self._loaded_all = True
        if refs is None:
            self._refs = tuple([*subname.split(".")])
        else:
            self._refs = tuple([*refs, *subname.split(".")])

        # Imitate the name attribute like a callable should
        self.__name__ = self._subname

    def __getstate__(self):
        sub = Sub.__getstate__(self)
        # These are all the args required to recreate the ReverseSub with hub.pop.sub.dynamic()
        sub["resolver"] = self._resolver
        sub["context"] = self._context
        # These are for when the reverse sub is wrapped directly
        sub["refs"] = self._refs
        sub["hub"] = self._hub
        return sub

    def __setstate__(self, state: Dict):
        """
        Sometimes a reverse sub will be called by proxy and can be restored by a hub.
        Sometimes it is called directly and the ReverseSub needs to have everything inside it.
        """
        self.__init__(
            hub=state.pop("hub"),
            subname=state["subname"],
            resolver=state.pop("resolver"),
            context=state.pop("context"),
            refs=state.pop("refs"),
        )
        Sub.__setstate__(self, state)

    def __getattr__(self, item: str):
        """
        Build a reference based on the namespace attrs until a reverse sub is called
        """
        if item.startswith("_"):
            return self.__getattribute__(item)

        return ReverseSub(
            hub=self._hub,
            subname=item,
            # Share a root
            root=self._root,
            resolver=self._resolver,
            context=self._context,
            pypath=self._pypath,
            refs=self._refs,
            static=self._static,
            contracts_pypath=self._contracts_pypath,
            contracts_static=self._recursive_contracts_static,
            default_contracts=self._default_contracts,
            virtual=self._virtual,
            dyne_name=self._dyne_name,
            omit_start=self._omit_start,
            omit_end=self._omit_end,
            omit_func=self._omit_func,
            omit_class=self._omit_class,
            omit_vars=self._omit_vars,
            mod_basename=self._mod_basename,
            stop_on_failures=self._stop_on_failures,
            init=None,
            is_contract=self._is_contract,
            sub_virtual=self._sub_virtual,
            recursive_contracts_static=self._recursive_contracts_static,
            default_recursive_contracts=self._default_recursive_contracts,
        )

    def __iter__(self):
        return self

    def __next__(self):
        # Override the parent Sub's iterator
        raise StopIteration

    def __contains__(self, item):
        """
        Return true if the item's reference underneath this one is resolvable
        """
        try:
            if getattr(self, item)._resolve():
                return True
        except Exception:
            ...
        return False

    def _resolve(self) -> Callable or None:
        """
        Call the resolver and get back the external call represented by this sub's reference
        """
        ref = ".".join(self._refs)
        ret = self._resolver(ref, self._context)
        return ret

    def __call__(self, *args, **kwargs):
        func = self._resolve()

        if not func:
            ref = ".".join(self._refs)
            raise AttributeError(f"ReverseSub could not resolve the reference {ref}")
        if isinstance(func, pop.contract.Contracted):
            discovered_contracted_function = func
        else:
            discovered_contracted_function = pop.contract.create_contracted(
                hub=self._hub,
                contracts=getattr(self._contracts, "_loaded", {}).values(),
                func=func,
                ref=".".join(self._refs),
                name=func.__name__,
                implicit_hub=False,
            )

        if self._recursive_contracts and not all(
            rc in discovered_contracted_function.contracts
            for rc in self._recursive_contracts
        ):
            discovered_contracted_function.contracts.extend(
                self._recursive_contracts._loaded.values()
            )
            discovered_contracted_function._load_contracts()

        return discovered_contracted_function(*args, **kwargs)



"""
Routines to verify the working environment etc.
"""
import os

import pop.hub


def env(hub: "pop.hub.Hub"):
    """
    Verify that the directories specified in the system exist
    """
    for key in hub.opts:
        if key.endswith("_dir"):
            try:
                os.makedirs(hub.opts[key])
            except OSError:
                pass



"""
Contracts to enforce loader objects
"""
import asyncio
import functools
import inspect
import os
from collections import namedtuple
from types import ModuleType
from typing import Dict
from typing import Iterable
from typing import List

import pop.exc
import pop.hub
import pop.verify


class ContractedContext(
    namedtuple(
        "ContractedContext",
        ("func", "args", "kwargs", "signature", "ret", "cache", "ref"),
    )
):
    """
    Contracted function calling context
    """

    def __new__(
        cls,
        func: functools.partial,
        args: Iterable,
        kwargs: Dict,
        signature,
        ref,
        ret=None,
        cache=None,
    ):  # pylint: disable=too-many-arguments
        if cache is None:
            cache = {}
        return super().__new__(
            cls, func, list(args), kwargs, signature, ret, cache, ref
        )

    def get_argument(self, name):
        """
        Return the value corresponding to a function argument after binding the contract context
        argument and keyword arguments to the function signature.
        """
        return self.get_arguments()[name]

    def get_arguments(self):
        """
        Return a dictionary of all arguments that will be passed to the function and their
        values, including default arguments.
        """
        if "__bound_signature__" not in self.cache:
            try:
                self.cache["__bound_signature__"] = self.signature.bind(
                    *self.args, **self.kwargs
                )
            except TypeError as e:
                for frame in inspect.trace(0):
                    if frame.function == "bind" and frame.filename.endswith(
                        os.sep + "inspect.py"
                    ):
                        raise pop.exc.BindError(e)
                raise
            # Apply any default values from the signature
            self.cache["__bound_signature__"].apply_defaults()
        return self.cache["__bound_signature__"].arguments


def load_contract(
    contracts: List["Contracted"],
    default_contracts: List[str],
    mod: ModuleType,
    name: str,
) -> List:
    """
    return a Contract object loaded up
    Dynamically create the correct Contracted type
    :param contracts: Contracts functions to add to the sub
    :param default_contracts: The contracts that have been marked as defaults
    :param mod: A loader module
    :param name: The name of the module to get from the loader
    """
    raws = []
    if not contracts:
        return raws
    loaded_contracts = []
    if hasattr(contracts, name):
        loaded_contracts.append(name)
        raws.append(getattr(contracts, name))
    if hasattr(contracts, "init"):
        loaded_contracts.append("init")
        raws.append(getattr(contracts, "init"))
    if default_contracts:
        for contract in default_contracts:
            if contract in loaded_contracts:
                continue
            loaded_contracts.append(contract)
            raws.append(getattr(contracts, contract))
    if hasattr(mod, "__contracts__"):
        cnames = getattr(mod, "__contracts__")
        if not isinstance(cnames, (list, tuple)):
            cnames = cnames.split(",")
        for cname in cnames:
            if cname in contracts:
                if cname in loaded_contracts:
                    continue
                loaded_contracts.append(cname)
                raws.append(getattr(contracts, cname))
    return raws


class Wrapper:
    def __init__(self, func: functools.partial, ref: str, name: str):
        """
        :param func: The contracted function to call
        :param ref: The reference to the function on the hub
        :param name: An alias for the function
        """
        self.__dict__.update(
            getattr(func, "__dict__", {})
        )  # do this first so we later overwrite any conflicts
        self.func = func
        self.ref = ref
        self.__name__ = name
        try:
            self.signature = inspect.signature(self.func)
        except ValueError:
            # inspect.signature raises a ValueError only when a signature could not be provided.
            # Some callables may not be introspectable in certain implementations of Python.
            # For example, in CPython, some built-in functions defined in C provide no metadata about their arguments.
            self.signature = None
        self._sig_errors = []
        self.__wrapped__ = func

    def __call__(self, *args, **kwargs):
        self.func(*args, **kwargs)

    def __repr__(self):
        return "<{} func={}.{}>".format(
            self.__class__.__name__, self.func.__module__, self.__name__
        )


class Contracted(Wrapper):
    """
    This class wraps functions that have a contract associated with them
    and executes the contract routines
    """

    def __init__(
        self,
        hub: "pop.hub.Hub",
        contracts: List["pop.hub.Sub"],
        func: functools.partial,
        ref: str,
        name: str,
        implicit_hub: bool = True,
    ):
        """
        :param hub: The redistributed pop central hub
        :param contracts: Contracts functions to add to the sub
        :param func: The contracted function to call
        :param ref: The reference to the function on the hub
        :param name: An alias for the function
        :param implicit_hub: True if a hub should be implicitly injected into the "call" method
        """
        super().__init__(func, ref, name)
        self.hub = hub
        self.contracts = contracts or []
        self.implicit_hub = implicit_hub
        self._load_contracts()

    def _get_contracts_by_type(self, contract_type: str = "pre") -> List[Wrapper]:
        """
        :param contract_type: One of "call", "pre", "post", or "sig"
        """
        matches = []
        fn_contract_name = f"{contract_type}_{self.__name__}"
        for contract in self.contracts:
            if hasattr(contract, fn_contract_name):
                matches.append(getattr(contract, fn_contract_name))
            if hasattr(contract, contract_type):
                matches.append(getattr(contract, contract_type))

        return matches

    def _load_contracts(self):
        # TODO:
        # if Contracted - only allow regular pre/post
        # if ContractedAsync - allow coroutines and functions
        # if ContractedAsyncGen - allow coroutines and functions

        self.contract_functions = {
            "pre": self._get_contracts_by_type("pre"),
            "call": self._get_contracts_by_type("call")[:1],
            "post": self._get_contracts_by_type("post"),
        }
        # TODO: write test for stack-like behavior (reverse "pre")
        self._has_contracts = sum(len(l) for l in self.contract_functions.values()) > 0

    def __call__(self, *args, **kwargs):
        if self.implicit_hub:
            args = (self.hub,) + args
        if not self._has_contracts:
            return self.func(*args, **kwargs)
        contract_context = ContractedContext(
            self.func, args, kwargs, self.signature, self.ref
        )

        for fn in self.contract_functions["pre"]:
            fn(contract_context)
        if self.contract_functions["call"]:
            ret = self.contract_functions["call"][0](contract_context)
        else:
            ret = self.func(*contract_context.args, **contract_context.kwargs)
        for fn in self.contract_functions["post"]:
            post_ret = fn(contract_context._replace(ret=ret))
            if post_ret is not None:
                ret = post_ret

        return ret

    def __getstate__(self):
        return dict(
            hub=self.hub,
            ref=self.ref,
            name=self.__name__,
            implicit_hub=self.implicit_hub,
            contracts=self.contracts,
        )

    def __setstate__(self, state):
        hub = state["hub"]
        ref = state["ref"]
        name = state["name"]
        func = hub[ref][name].func
        self.__init__(
            hub=hub,
            func=func,
            ref=ref,
            name=name,
            implicit_hub=state["implicit_hub"],
            contracts=state["contracts"],
        )


class ContractedAsyncGen(Contracted):
    async def __call__(self, *args, **kwargs):
        if self.implicit_hub:
            args = (self.hub,) + args
        if not self._has_contracts:
            async for chunk in self.func(*args, **kwargs):
                yield chunk
            return
        contract_context = ContractedContext(
            self.func, args, kwargs, self.signature, self.ref
        )

        for fn in self.contract_functions["pre"]:
            pre_ret = fn(contract_context)
            if asyncio.iscoroutine(pre_ret):
                await pre_ret
        chunk = None
        if self.contract_functions["call"]:
            async for chunk in self.contract_functions["call"][0](contract_context):
                yield chunk
        else:
            async for chunk in self.func(
                *contract_context.args, **contract_context.kwargs
            ):
                yield chunk
        ret = chunk
        for fn in self.contract_functions["post"]:
            if isinstance(fn, ContractedAsync):
                post_ret = await fn(contract_context._replace(ret=ret))
            else:
                post_ret = fn(contract_context._replace(ret=ret))
            if post_ret is not None:
                ret = post_ret


class ContractedAsync(Contracted):
    async def __call__(self, *args, **kwargs):
        if self.implicit_hub:
            args = (self.hub,) + args
        if not self._has_contracts:
            return await self.func(*args, **kwargs)
        contract_context = ContractedContext(
            self.func, args, kwargs, self.signature, self.ref
        )

        for fn in self.contract_functions["pre"]:
            pre_ret = fn(contract_context)
            if asyncio.iscoroutine(pre_ret):
                await pre_ret
        if self.contract_functions["call"]:
            ret = await self.contract_functions["call"][0](contract_context)
        else:
            ret = await self.func(*contract_context.args, **contract_context.kwargs)
        for fn in self.contract_functions["post"]:
            if isinstance(fn, ContractedAsync):
                post_ret = await fn(contract_context._replace(ret=ret))
            else:
                post_ret = fn(contract_context._replace(ret=ret))
            if post_ret is not None:
                ret = post_ret

        return ret


def create_contracted(
    hub: "pop.hub.Hub",
    contracts: List["pop.loader.LoadedMod"],
    func: functools.partial,
    ref: str,
    name: str,
    implicit_hub: bool = True,
) -> Contracted:
    """
    Dynamically create the correct Contracted type
    :param hub: The redistributed pop central hub
    :param contracts: Contracts functions to add to the sub
    :param func: The contracted function to call
    :param ref: The reference to the function on the hub
    :param name: The name of the module to get from the loader
    :param implicit_hub: True if a hub should be implicitly injected into the "call" method
    """
    if asyncio.iscoroutinefunction(func):
        return ContractedAsync(hub, contracts, func, ref, name, implicit_hub)
    elif inspect.isasyncgenfunction(func):
        return ContractedAsyncGen(hub, contracts, func, ref, name, implicit_hub)
    else:
        return Contracted(hub, contracts, func, ref, name, implicit_hub)


class Partial:
    """
    A partial function that can be used for multiprocessing
    """

    def __init__(self, synchronous_function: Contracted, *args, **kwargs):
        self.func = synchronous_function
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        ret = self.func(*self.args, **self.kwargs)
        # We can't pickle a coroutine, so async functions need to finish here
        # if this is executed in a forked process, we can't guarantee a loop exists
        if asyncio.iscoroutine(ret):
            loop = asyncio.new_event_loop()
            try:
                while asyncio.iscoroutine(ret):
                    ret = loop.run_until_complete(ret)
            finally:
                loop.close()
        return ret

    def __getstate__(self):
        return {
            "func": self.func,
            "args": self.args,
            "kwargs": self.kwargs,
        }

    def __setstate__(self, state):
        self.func = state["func"]
        self.args = state["args"]
        self.kwargs = state["kwargs"]



from typing import Any
from typing import Dict

import pop.hub


def get(hub: "pop.hub.Hub") -> Dict[str, Any]:
    """
    Retrive the dynamic dirs data for this hub, if dynamic dirs have not been
    gathered yet then gather it.
    """
    if not hub._dscan:
        hub._scan_dynamic()
    return hub._dynamic


def refresh(hub: "pop.hub.Hub"):
    """
    Refresh the dynamic dirs
    """
    hub._scan_dynamic()



import collections
import collections.abc as abc
import copy
import inspect
import logging
import sys
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List

import pop.contract as contract
import pop.hub

log = logging.getLogger(__name__)

__func_alias__ = {
    "immutable_namespaced_map": "imap",
    "mutable_namespaced_map": "map",
    "owner_writeable_namespaced_map": "omap",
}


def immutable_namespaced_map(
    hub: "pop.hub.Hub", init: Dict[str, Any], **kwargs
) -> abc.MutableMapping:
    return IMAP(init_=init, **kwargs)


class IMAP(abc.Mapping):
    """
    An abstract base class that implements the interface of a `dict` but is immutable.
    Items can be retrieved via namespacing.
    No values can be changed after initialization
    """

    def __init__(self, init_: Dict[str, Any], **c_kwargs):
        """
        :param init_: A dictionary from which to inherit data
        """
        init_.update(**c_kwargs)
        values = {}
        for k, v in init_.items():
            if isinstance(v, Dict):
                values[k] = IMAP(init_=v)
            elif isinstance(v, (tuple, int, str, bytes)):
                values[k] = v
            elif isinstance(v, Iterable):
                values[k] = tuple(v)
            else:
                values[k] = v
        # __setattr__ is borked (on purpose) so we have to call it from super() right here
        super().__setattr__("_IMAP__store", values)
        log.debug("Initialized immutable namespaced map")

    def __setattr__(self, k: str, v: Any):
        raise TypeError(
            f"{self.__class__.__name__} does not support attribute assignment"
        )

    def __getattr__(self, k: str):
        if k.startswith("_"):
            return super().__getattribute__(k)
        else:
            return self.__store[k]

    def __getitem__(self, k: str) -> Any:
        return self.__store[k]

    def __contains__(self, k: str) -> bool:
        return k in self.__store

    def __iter__(self):
        return iter(self.__store)

    def __len__(self) -> int:
        return len(self.__store.keys())

    def __copy__(self) -> Dict[str, Any]:
        ret = {}
        # Unpack IMAP items so that it's turtles all the way down
        for k, v in self.__store.items():
            if isinstance(v, IMAP):
                ret[k] = v.__copy__()
            else:
                ret[k] = v
        return ret

    def __repr__(self):
        return repr(copy.copy(self))


def mutable_namespaced_map(hub: "pop.hub.Hub", dict_: Dict[str, Any] = None) -> "MAP":
    return MAP(dict_=dict_)


class WriteLockError(Exception):
    pass


class MAP(abc.MutableMapping):
    """
    MAP is a key-value store that allows for setting/getting
    by either dot or dictionary lookup notation ('.' or '[k]')

    Sub-keys will be created on assignment:

      `map.foo.bar.baz = True` will auto-create foo and bar as MAPs

    while doing

      `map.foo.bar.baz` before assignment will not create foo, bar or baz.

    :param dict_: similar to dict(dict_), initialize using dict_
    """

    def __init__(self, dict_: Dict[str, Any] = None, parent: "MAP" = None):
        self.__dict__["_store"] = {}
        self.__dict__["_parent"] = parent
        if dict_:
            # Existing dictionaries might have properties that need wrapped as well
            self.update(dict_)

    def _set(self, k: str, v: Any):
        if k.startswith("_"):
            raise AttributeError("Cannot store values beginning with '_'")

        if isinstance(v, dict):
            # Cast all nested dict values as MAP so they get it's benefits as well
            v = self.__class__(dict_=v, parent=self)
            self._store[k] = v
        else:
            self._store[k] = v

    def _get(self, k: str, create: bool = False):
        if k.startswith("_"):
            return super().__getattribute__(k)
        try:
            if k not in self._store:
                if not create:
                    return UninitializedValue([k], self)
                self._set(k, self.__class__())
            return self._store[k]
        except Exception as e:
            raise AttributeError(*e.args)

    def get(self, k: str, default: Any = None) -> Any:
        if k in self._store:
            return getattr(self, k)
        else:
            return default

    def __setitem__(self, k: str, v: Any):
        self._set(k, v)

    def __delitem__(self, k: str):
        """
        Cleanup method required by abc.ABC
        """
        if k in self._store:
            del self._store[k]

    def __delattr__(self, k: str):
        self.__delitem__(k)

    def __getitem__(self, k: str) -> Any:
        return self._get(k)

    def __getattr__(self, k: str) -> Any:
        return self._get(k)

    def __setattr__(self, k: str, v: Any):
        self._set(k, v)

    def __contains__(self, k: str) -> Any:
        return k in self._store

    def __len__(self) -> int:
        return len(self._store)

    def __iter__(self) -> Iterator[Any]:
        return iter(self._store)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self)})"

    def __str__(self) -> str:
        return str(self._store)

    def __copy__(self) -> Dict[str, Any]:
        # The copy will be a dictionary, mangle it all you want
        ret = {}
        for k, v in self._store.items():
            if isinstance(v, MAP):
                ret[k] = v.__copy__()
            else:
                ret[k] = v
        return ret


class UninitializedValueError(Exception):
    pass


class UninitializedValue:
    """
    We want you to be able to create data on the MAP by doing something like:

        MAP.foo.bar = True

    However, we want uses of un-initialized values to blow up, not auto-create.

    We build up a potentially pending write using a special object that can only be written.
    Any other use should blow up.

    """

    def __init__(self, path: List[str], map_: MAP):
        # The path I'm looking up on the map
        self.__dict__["_path"] = path
        self.__dict__["_map"] = map_

    def _blowup(self):
        raise UninitializedValueError(
            f"Access of uninitialized value '{'.'.join(self.__dict__['_path'])}'"
        )

    def __getattribute__(self, item: str):
        if item == "get":
            pass
        elif not item.startswith("_"):  # return a new PendingWrite for regular lookups
            new_path = self._path[:]
            new_path.append(item)
            return UninitializedValue(new_path, self._map)
        elif item not in (
            "__class__",
            "__dict__",
            "__setattr__",
            "_blowup",
            "_map",
            "_path",
            "get",
        ):
            self._blowup()
        return super().__getattribute__(item)

    def __setattr__(self, k: str, v: Any):
        node = self._map
        for part in self._path:
            node = node._get(part, create=True)

        setattr(node, k, v)

    def __delattr__(self, k: str):
        pass

    def __getitem__(self, item: str):
        return getattr(self, item)

    def __setitem__(self, k: str, v: Any):
        setattr(self, k, v)

    def __delitem__(self, k: str):
        pass

    def get(self, k: str, default: Any = None) -> Any:
        return default

    def __bool__(self) -> bool:
        # Uninitialized values always evaluate to false
        return False

    def __contains__(self, item: str) -> bool:
        # If it contained anything it would be initialized
        return False

    def __hash__(self):
        self._blowup()

    def __dir__(self):
        self._blowup()

    def __str__(self):
        self._blowup()


def owner_writeable_namespaced_map(
    hub, dict_: Dict[str, Any] = None
) -> "OwnerWriteableMapping":
    return OwnerWriteableMapping(dict_=dict_)


def _stack_frames(relative_start=2):
    """
    Efficiently access stack frames.
    :param relative_start: Starting stack depth; The default, 2 is the parent of the
                           caller of stack_frames - the first function that may be unknown.
    :return: a stack frame
    """
    if hasattr(sys, "_getframe"):
        # implementation detail of CPython, speeds things up by 100x.
        frame = sys._getframe(relative_start)
        while frame:
            yield frame
            frame = frame.f_back
    else:
        for frame_info in inspect.stack(context=0)[relative_start:]:
            yield frame_info.frame


WriteLockInfo = collections.namedtuple("WriteLockInfo", ["val", "owner", "lineno"])


class OwnerWriteableMapping(MAP):
    """
    A MAP variant that is write-locked to the first Contracted function
    that writes to a given key (becoming the owner). Attempts to write
    to that key from other functions will receive a WriteLockError showing
    the owning Contracted function.
    """

    def __init__(
        self, dict_: Dict[str, Any] = None, parent: "OwnerWriteableMapping" = None
    ):
        super().__init__(dict_, parent)

    def _find_owner(self) -> (contract.Contracted, int):
        """
        Return the contracted responsible for assigning to this variable.
        Returns None if no such function exists.
        """
        for frame in _stack_frames(3):
            if isinstance(frame.f_locals.get("self"), contract.Contracted):
                contracted = frame.f_locals["self"]
                log.debug(f"Found contract '{contracted.__name__}'")
                break
            else:
                # find the lineno in the frame *before* our Contracted (the function called)
                lineno = frame.f_lineno
        else:
            # not found
            contracted = None
            lineno = -1

        return contracted, lineno

    def _set(self, k: str, v: Any):
        owner, lineno = self._find_owner()
        cur = self._store.get(k)
        if cur is None or cur.owner is owner:
            if isinstance(v, abc.Mapping):
                v = self.__class__(dict_=v, parent=self)
            elif isinstance(v, Iterable) and not isinstance(
                v, (tuple, str, bytes, UninitializedValue)
            ):
                v = tuple(v)  # Lists, sets, and other iterables become immutable
            super()._set(k, WriteLockInfo(v, owner, lineno))
        else:
            file = inspect.getsourcefile(cur.owner.func)
            raise WriteLockError(
                f"'{k}' was previously assigned by '{cur.owner.__name__}' ({file}:{cur.lineno})"
            )

    def _get(self, k: str, create: bool = False) -> Any:
        v = super()._get(k, create)
        if not isinstance(v, UninitializedValue):
            v = v.val
        return v

    def __str__(self) -> str:
        return str(self._dict())

    def _dict(self):
        vals = {}
        for k, v in self._store.items():
            if isinstance(v.val, self.__class__):
                vals[k] = v.val._dict()
            else:
                vals[k] = v.val
        return vals
